//
//  FileProviderExtension.m
//  DemoSomeFiles
//
//  Created by Amy Worrall on 04/09/2017.
//  Copyright © 2017 Amy Worrall. All rights reserved.
//

#import "FileProviderExtension.h"
#import "FileProviderItem.h"
#import "FileDBProviderProtocol.h"
#import "WFTopLevelFoldersEnumerator.h"
#import "WFFolderContentsEnumerator.h"

@interface FileProviderExtension ()<FileDBProvider>

@property (nonatomic, readonly, strong) NSFileManager *fileManager;

@end

@implementation FileProviderExtension
{
  NSURL *_databaseURL;
}

- (instancetype)init {
    if (self = [super init]) {
      _fileManager = [[NSFileManager alloc] init];
      _databaseURL = [NSURL fileURLWithPath:[NSSearchPathForDirectoriesInDomains(NSDocumentDirectory, NSUserDomainMask, YES)[0] stringByAppendingPathComponent:@"fileDB"]];
    }
    return self;
}

- (nullable NSFileProviderItem)itemForIdentifier:(NSFileProviderItemIdentifier)identifier error:(NSError * _Nullable *)error {
  if ([identifier isEqualToString:NSFileProviderRootContainerItemIdentifier]) {
    return nil;
  }

  return [self entryForFBID:identifier];
}

- (nullable NSURL *)URLForItemWithPersistentIdentifier:(NSFileProviderItemIdentifier)identifier {
    // resolve the given identifier to a file on disk
    NSFileProviderItem item = [self itemForIdentifier:identifier error:NULL];
    if (!item) {
        return nil;
    }
    
    // in this implementation, all paths are structured as <base storage directory>/<item identifier>/<item file name>
    NSFileProviderManager *manager = [NSFileProviderManager defaultManager];
    NSURL *perItemDirectory = [manager.documentStorageURL URLByAppendingPathComponent:identifier isDirectory:YES];
    
    return [perItemDirectory URLByAppendingPathComponent:item.filename isDirectory:NO];
}

- (nullable NSFileProviderItemIdentifier)persistentIdentifierForItemAtURL:(NSURL *)url {
    // resolve the given URL to a persistent identifier using a database
    NSArray <NSString *> *pathComponents = [url pathComponents];
    
    // exploit the fact that the path structure has been defined as
    // <base storage directory>/<item identifier>/<item file name> above
    NSParameterAssert(pathComponents.count > 2);
    
    return pathComponents[pathComponents.count - 2];
}

- (void)startProvidingItemAtURL:(NSURL *)url completionHandler:(void (^)(NSError *))completionHandler {
  [[NSFileManager defaultManager] createDirectoryAtURL:[url URLByDeletingLastPathComponent] withIntermediateDirectories:YES attributes:nil error:nil];
  [[NSFileManager defaultManager] createFileAtPath:[url path] contents:nil attributes:nil];

  FileProviderItem *fileItem = [self entryForFBID:[self persistentIdentifierForItemAtURL:url]];

  NSURL *aURL = [NSURL URLWithString:fileItem.downloadURL];

  NSURLSessionDownloadTask *downloadItemTask =
  [[NSURLSession sharedSession]
   downloadTaskWithURL:aURL completionHandler:^(NSURL *location, NSURLResponse *response, NSError *error) {
     NSError *newError;
     [[NSFileManager defaultManager] removeItemAtURL:url error:&error];
     [[NSFileManager defaultManager] moveItemAtURL:location toURL:url error:&newError];
     completionHandler(nil);
   }];

  [downloadItemTask resume];
}


- (void)itemChangedAtURL:(NSURL *)url {
    // Called at some point after the file has changed; the provider may then trigger an upload
    
    /* TODO:
     - mark file at <url> as needing an update in the model
     - if there are existing NSURLSessionTasks uploading this file, cancel them
     - create a fresh background NSURLSessionTask and schedule it to upload the current modifications
     - register the NSURLSessionTask with NSFileProviderManager to provide progress updates
     */
}

- (void)stopProvidingItemAtURL:(NSURL *)url {
    // Called after the last claim to the file has been released. At this point, it is safe for the file provider to remove the content file.
    
    // TODO: look up whether the file has local changes
    BOOL fileHasLocalChanges = NO;
    
    if (!fileHasLocalChanges) {
        // remove the existing file to free up space
        [[NSFileManager defaultManager] removeItemAtURL:url error:NULL];
        
        // write out a placeholder to facilitate future property lookups
        [self providePlaceholderAtURL:url completionHandler:^(NSError * __nullable error) {
            // TODO: handle any error, do any necessary cleanup
        }];
    }
}

- (void)providePlaceholderAtURL:(NSURL *)url completionHandler:(void (^)(NSError * _Nullable))completionHandler
{
  FileProviderItem *entry = [self entryForFBID:[self persistentIdentifierForItemAtURL:url]];

  NSError *error;
  NSURL *place = [NSFileProviderManager placeholderURLForURL:url];
  [[NSFileManager defaultManager] createDirectoryAtURL:[place URLByDeletingLastPathComponent] withIntermediateDirectories:YES attributes:nil error:nil];
  [NSFileProviderManager writePlaceholderAtURL:place withMetadata:entry error:&error];

  if (error) {
    NSLog(@"Write error: %@", error);
  }

  if (completionHandler) {
    completionHandler(error);
  }
}

#pragma mark - Actions

/* TODO: implement the actions for items here
 each of the actions follows the same pattern:
 - make a note of the change in the local model
 - schedule a server request as a background task to inform the server of the change
 - call the completion block with the modified item in its post-modification state
 */

#pragma mark - Enumeration

- (nullable id<NSFileProviderEnumerator>)enumeratorForContainerItemIdentifier:(NSFileProviderItemIdentifier)containerItemIdentifier error:(NSError **)error {
    id<NSFileProviderEnumerator> enumerator = nil;
    if ([containerItemIdentifier isEqualToString:NSFileProviderRootContainerItemIdentifier]) {
        // TODO: instantiate an enumerator for the container root
      WFTopLevelFoldersEnumerator *foldersEnumerator = [[WFTopLevelFoldersEnumerator alloc] initWithEnumeratedItemIdentifier:NSFileProviderRootContainerItemIdentifier];
      foldersEnumerator.delegate = self;
      enumerator = foldersEnumerator;

    } else if ([containerItemIdentifier isEqualToString:NSFileProviderWorkingSetContainerItemIdentifier]) {
        // TODO: instantiate an enumerator for the working set
    } else {
        // TODO: determine if the item is a directory or a file
        // - for a directory, instantiate an enumerator of its subitems
        // - for a file, instantiate an enumerator that observes changes to the file
      FileProviderItem *entry = [self entryForFBID:containerItemIdentifier];
      if (entry.isFolder) {
        // it's a directory
        WFFolderContentsEnumerator *filesEnumerator = [[WFFolderContentsEnumerator alloc] initWithEnumeratedItemIdentifier:containerItemIdentifier];
        filesEnumerator.delegate = self;
        enumerator = filesEnumerator;
      } else {
        // it's a file
        // TODO
      }

    }
    
    return enumerator;
}

- (NSProgress *)fetchThumbnailsForItemIdentifiers:(NSArray<NSFileProviderItemIdentifier> *)itemIdentifiers requestedSize:(CGSize)size perThumbnailCompletionHandler:(void (^)(NSFileProviderItemIdentifier _Nonnull, NSData * _Nullable, NSError * _Nullable))perThumbnailCompletionHandler completionHandler:(void (^)(NSError * _Nullable))completionHandler
{
  NSProgress *progress = [NSProgress progressWithTotalUnitCount:itemIdentifiers.count];

  __block NSUInteger numProvidersLeft = itemIdentifiers.count;
  for (NSFileProviderItemIdentifier identifier in itemIdentifiers) {
    FileProviderItem *fileItem = [self entryForFBID:identifier];

    // our web service names its thumbnails based on the file's url
    NSURL *thumbnailURL = [NSURL URLWithString:[NSString stringWithFormat:@"%@.thumb.png", fileItem.downloadURL]];

    NSURLSessionDataTask *thumbnailTask =
    [[NSURLSession sharedSession]
     dataTaskWithURL:thumbnailURL completionHandler:^(NSData * _Nullable data, NSURLResponse * _Nullable response, NSError * _Nullable error) {
       if (data) {
         perThumbnailCompletionHandler(identifier, data, error);
       } else {
         perThumbnailCompletionHandler(identifier, nil, nil);
       }
       numProvidersLeft--;
       if (numProvidersLeft == 0) {
         completionHandler(nil);
       }
     }];

    [thumbnailTask resume];
  }

  return progress;
}


#pragma mark - Higher level files DB operations

- (FileProviderItem *)entryForFBID:(NSString *)fbid
{
  return [FileProviderItem fromDictionary:[self filesDB][fbid]];
}

- (void)setEntry:(FileProviderItem *)entry forFBID:(NSString *)fbid;
{
  NSMutableDictionary *dict = [[self filesDB] mutableCopy];
  dict[fbid] = [entry toDictionary];
  [self saveFilesDB:dict];
}

#pragma mark - Database

- (NSDictionary *)filesDB
{
  NSDictionary *dict = [NSDictionary dictionaryWithContentsOfURL:_databaseURL];
  if (!dict) {
    dict = @{};
  }
  return dict;
}

- (void)saveFilesDB:(NSDictionary *)newDB
{
  [newDB writeToURL:_databaseURL atomically:YES];
}


@end

